'Filename   SWRMeter_Mega32_201.bas
'Version    2.01
'Date       2014-03-05
'Purpose    SWR and Power meter with
'           - mathematical approximation of the diode U/I-characteristic
'             to calculate the 'real' HF voltage from measured diode DC voltage
'             valid for Schotty BAT43.
'             For other diodes pls. measure the characteristic and
'             fit the data to the function y = a * x^b + 1, x = DC voltage
'             with the help of the Excel function RKP.
'           - display of SWR & RF foreward/reverse power
'           - Fine tuning can be done with gain adjustment on OpAmps.
'             For reliable power calibration pls. use a JFET/CMOS opamp
'             with adjustable 1.5 ... 2x gain, e.g. TLC272.
'           - Choose toroid turns of tandem match coupler fitting max. RF power
'             Change (see below): bytTurns = 9, fits 20W
'           - SWR bridge data are set in SWR_bridge_data.bas
'             sngADCRef = 4982     ADC supply/reference voltage AVCC (mV)
'             bytTurns = 9          Power calibration factor
'                                   =turns of tandem match coupler toroids
'             sngDiodeA = 0.251607  Diode fitparameter "a"
'             sngDiodeB = -0.857977 Diode fitparameter "b"
'             sngDiodeC = 1         Diode fitparameter "c"
'             sngDiodeLow = 0.02    lower Diode detection limit (V)
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega32 (ATmega8 would be sufficient, 67% flash used)
'           Crystal 16 MHz
'           Ports PA0, PA1 (ADC0, ADC1): SWR Fwd, Rev
'           Port B0: Jumper or switch eff./peak HF power
'           Port B1: LED_Power eff./peak HF power, lit when Port B0 is grounded
'           Port B2: LED_SWR is lit when SWR > 3
'           Ports D.0 ... D5: 2x16 Text LCD
'           Customizing option, future use:
'           Port C0: Right key
'           Port C1: Up key
'           Port C2: Left key
'Compiler   BASCOM-AVR 2.0.7.3
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'           Flash used: 12,150 bytes (74%, ATmega16, 37%, ATmega32)
'Edit history
'V1.00      2012-08-04, first release
'V1.10      2013-09-11
'           BASCOM "GETADC" replaced by ADC register handling (faster)
'           Added: Jumper/switch for HF eff/peak power display
'           Added: LED_Power to show HF eff/peak power setting
'           Added: LED_SWR: SWR > 3 alert
'           SWR bridge data set and included via SWR_Bridge_Data.bas
'           Prepared for future setup option, not implemented yet
'V2.00      2013-11-02
'           Added: Coupler data setup, 2 couplers may be selected
'V2.01      2014-03-05
'           Repetitive ADC runs increased from 3 to 7
'           New: Voltage measurement calibration factors for the 2 couplers/units
'           EEPROM storage for the 2 couplers changed
'==============================================================================

'FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'PonyProg does'nt know the ATmega1284P
'In ATMEGA-Data sheets: Bit = 1 -> not set
'                       Bit = 0 -> set
'BODLEVEL   disabled  (BODLEVEL=111)
'OCDEN      [ ]
'JTAGEN     [ ] (disabled)
'SPIEN      [x] Do not change!
'WDTON      [ ]
'EESAVE     [ ]
'BOOTSZ     Boot Flash size=512 words (minimal boot size)
'BOOTRST    [ ]
'CKDIV8     [ ]
'CKOUT      [ ]
'SUT_CKSEL  Ext. Crystal Osc. 8.0- MHz, Start-up time 16k + 65ms

'$regfile = "m16def.dat"                           'ATMega16
$regfile = "m32def.dat"                           'ATMega32

$crystal = 16000000                               '16 MHz xtal
$hwstack = 64
$swstack = 64
$framesize = 64

'Customizing variables for ADC measurement ------------------------------------
'System impedance sngImp, usually sngImp = 50 Ohm
'Max. SWR alert sngSWRMax, LED_SWR is lit
Const sngImp = 50                                 'System impedance
Const sngSWRMax = 3                               'Max. SWR: LED_SWR is lit

'Diode response parameters ----------------------------------------------------
'Best fit for Schottky diode correction function
'CF = a * DC^b + c
'DC = measured rectified DC voltage
'HF = CF * DC, best fit for the "real" HF voltage
Dim sngDiodeLow As Single                         'lower Diode detection limit (V)
Dim sngDiodeA As Single                           'Parameter "a"
Dim sngDiodeB As Single                           'Parameter "b"
Dim sngDiodeC As Single                           'Parameter "c"
Dim sngCalUnit As Single                          'Unit voltage calibr. factor

'ADC & selection coupler1/coupler3 on Port A ----------------------------------
'ADC0                      PA0                    'Forward voltage
'ADC1                      PA1                    'Reverse voltage
'ADC2 to ADC5              PA2 to PA5             'Not used
'Jmp_Coupler2              PA6                    '=1: Coupler1, =0: Coupler2
'LED_Coupler2              PA7                    'LED lit: Coupler2 selected
DdrA = &B10000000                                 'PortA = input except PA7=out
PortA = &B01111100                                'Pullup PA2...PA6
Jmp_Coupler2 Alias PINA.6                         'Key coupler selection
LED_Coupler2 Alias PORTA.7                        'LED coupler2 selected

'Configure in/out on Port B ---------------------------------------------------
'J1   Jumper RF power         PB0
'LED_Power RF power eff/peak  PB1, lit: eff. power
'LED_SWR SWR > 3              PB2, lit: SWR > 3
DDRB = &B00000110                                 'All input exept PB1, PB2
PortB = &B00011001                                'Pullup PB0, PB3, PB4
Jmp_Power Alias PinB.0                            'Jumper RF eff/peak
LED_Power Alias PortB.1                           'Output LED_Power @ PortB.1
LED_SWR Alias PortB.2                             'Output LED_SWR @ PortB.2

'Configure Customizing option switches on Port C ------------------------------
DdrC = &B00000000                                 'PortC = input
PortC = &B11111111                                'Pullup PC0...PC7
Key_port Alias PinC                               'Input-Port Keys
Jmp_Setup Alias PINC.0                            'Jumper customize
Const Key_Mode = 1                                'Key mode        @ PINC.1
Const Key_Left = 2                                'Key digit left  @ PinC.2
Const Key_Up = 3                                  'Key numbers up  @ PINC.3
Const Key_Right = 4                               'Key digit right @ PinC.4

'Configure LCD on Port D -----------------------------------------------------
'PD0  LCD      RS
'PD1  LCD      E
'PD2  LCD      D4
'PD3  LCD      D5
'PD4  LCD      D6
'PD5  LCD      D7
'PD6  Not used
'PD7  Not used
DDRD.6 = 0                                        'PD.6 input
DDRD.7 = 0                                        'PD.7 input
PORTD.6 = 1                                       'PD.6 pullup
PORTD.7 = 1                                       'PD.7 pullup
Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortD.5 , Db6 = PortD.4 , Db5 = PortD.3
Config Lcdpin = Pin , Db4 = PortD.2 , E = PortD.1 , Rs = PortD.0
Config Lcdbus = 4

Const bytRowPwr = 1                               'Row for RF power display
Const bytRowSWR = 2                               'Row for SWR display
Const bytChar = 16                                'Characters per LCD line
Const bytDots = 5                                 'bar dots per character
Const bytBarLen = 12                              'Bar length (characters)

'Define display bars ----------------------------------------------------------
Deflcdchar 0 , 16 , 16 , 16 , 16 , 16 , 16 , 16 , 32       'chr(0): 1 bar
Deflcdchar 1 , 24 , 24 , 24 , 24 , 24 , 24 , 24 , 32       'chr(1): 2 bars
Deflcdchar 2 , 28 , 28 , 28 , 28 , 28 , 28 , 28 , 32       'chr(2): 3 bars
Deflcdchar 3 , 30 , 30 , 30 , 30 , 30 , 30 , 30 , 32       'chr(3): 4 bars
Deflcdchar 4 , 31 , 31 , 31 , 31 , 31 , 31 , 31 , 32       'chr(4): 5 bars (full)

'Variables program control ----------------------------------------------------
Dim bytMode As Byte                               'Index of bytMode_state
Dim bytLong As Byte
Dim bitSetup As Bit
Dim bytSave As Byte
Dim bytPos As Byte
Dim bytCursorPos As Byte
Dim bytCursorFirst As Byte                        'First LCD cursor position
Dim bytNumChar As Byte                            'Number of LCD characters
Dim bytSignPos As Byte                            'Position of number sign
Dim bytPointPos As Byte                           'Position of number dec. point

'Common Variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte
Dim wrdTmp0 As Word
Dim wrdTmp1 As Word
Dim wrdTmp2 As Word
Dim sngTmp0 As Single
Dim sngTmp1 As Single
Dim strTmp1 As String * 1
Dim strTmp6 As String * 6
Dim strDis As String * 16                         'Text to be dispayed on LCD
Dim bytDis(16) As Byte At strDis Overlay          'the same as byte array
Dim strTmp10 As String * 10
Dim bytTmp10(10) As Byte At strTmp10 Overlay

'Variables Keys ---------------------------------------------------------------
Dim bytIwr0 As Byte
Dim bytKey_ct0 As Byte
Dim bytKey_ct1 As Byte
Dim bytKey_state As Byte
Dim bytKey_press As Byte
Dim bytKey_save As Byte
Dim bytTimer0_runs As Byte
Dim bitKey_Long As Bit
Const KeyCounter = 60                             '~ 1 sec, pressed long

'ADC Variables ----------------------------------------------------------------
Dim bytADMUX As Byte
Dim wrdADC As Word
Dim bytADCLo As Byte At wrdADC Overlay
Dim bytADCHi As Byte At wrdADC + 1 Overlay
Dim bitADCdone As Bit                             '1= all ADC channels complete
Dim bitADCPWR As Bit                              '=1: Power accumulation done
Dim bytADCRun As Byte                             'Current ADC run for mean value
Dim wrdADCVal(2) As Word                          '1=Forward, 2=Reverse
Dim bytADChannel As Byte                          '0=Forward, 1=Reverse
Dim bytADCValIdx As Byte                          'Current ADC value index
Dim bytPWRerr As Byte
Const bytADCRep = 7                               'Repetitive ADC runs for mean value
Const bytADCUsed = 1                              'Number of used ADC channels - 1

'SWR & power Variables --------------------------------------------------------
Dim bytTurns As Byte                              'Power calibration factor
Dim sngADCRef As Single                           'DC supply/reference voltage AVCC (V)
Dim sngADCFwd As Single                           'ADC forward voltage
Dim sngADCRev As Single                           'ADC reverse voltage
Dim sngFwd As Single                              'Mean ADC forward voltage
Dim sngRev As Single                              'Mean ADC reverse voltage
Dim sngSWR As Single                              'SWR
Dim sngADCCal As Single                           '10 bit ADC calibration
Dim sngSWRBarCal As Single                        'SWR bar scale factor
Dim bytUnit As Byte                               'Bridge units 1 or 2
Const sngSWRBarMax = 5                            'Max SWR for bargraph

'EEPROM data ------------------------------------------------------------------
Dim wrdAddr As Word
Dim wrdEEPROM As Word                             'Word variable to/from EEPROM
Dim bytWrdEEP(2) As Byte At wrdEEPROM Overlay
Dim sngEEPROM As Single                           'Single to/from EEPROM
Dim bytSngEEP(4) As Byte At sngEEPROM Overlay

'Declare subs & functions -----------------------------------------------------
Declare Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
Declare Function KeyPressLong(byVal KeyNum As Byte) As Byte
Declare Sub GetADCData
Declare Sub ADCalibrate
Declare Function DiodeResponse(byVal sngDC As Single) As Single
Declare Sub ShowPower
Declare Sub ShowSWR
Declare Sub SetValue(byVal bytRow As Byte , byVal bytType As Byte , byVal intMult As Integer)
Declare Sub CheckEEPROM
Declare Sub EEPROMReadWrd(byRef wrdWord As Word)
Declare Sub EEPROMWriteWrd(byRef wrdWord As Word)
Declare Sub EEPROMReadSng(byRef sngSingle As Single)
Declare Sub EEPROMWriteSng(byRef sngSingle As Single)

'Configure Timer0-Interrupt for Keys (Timer0 = 8bit, 2^8=256 counts) ----------
'Overflow time: Overflow-Counts * Prescale / crystal frequency
'= 255 * 1024 / 16.000.000 = 16,3 msec = ~ 61 Hz (crystal 16 MHz)
'Possible Prescales: 8, 64, 256, 1024, here 1024
Config Timer0 = Timer , Prescale = 1024
Const PresetTimer0 = 56                           'Initialize Timer0 (shorten time, >0)
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr

'Configure ADC, Reference = AVCC (ATmega8/16/32) ------------------------------
bytADMUX = &B01000000                             'Reference = AVCC, ADLAR=0, ADC0
Admux = bytADMUX                                  'ADC Multiplexer Selection Register
'                                                  Bit 7 , 6 : REFS1 , REFS0
'                                                  00: Reference = AREF
'                                                  01: Reference = AVCC
'                                                  10: Int. 1.10V Reference
'                                                  11: Int. 2.56V Reference
'                                                  Bit 5, ADLAR=0: data right adjusted
'                                                              =1: data left adjusted
'                                                  Bit 4-0, MUX4...MUX0
'                                                  =0000: ADC0 is input
'                                                  =0001: ADC1 is input
'                                                  =0010: ADC2 is input
'                                                  =0011: ADC3 is input
'                                                  =0100: ADC4 is input
'                                                  =0101: ADC5 is input
'                                                  =0110: ADC6 is input
'                                                  =0111: ADC7 is input
ADCSRA = &B10001111                               'Enable ADC, div. factor 128
'                                                  ADC Control & Status Register A
'                                                  Bit 7, ADEN = 1: enable ADC
'                                                  Bit 6, ADSC = 1: Start ADC
'                                                  Bit 5, ADATE = 1: Auto trigger enabled
'                                                  Bit 4, ADIF = 1 when conversion completed
'                                                  Bit 3, ADIE = 1: Conversion completion interrupt activated
'                                                  Bit 2-0, ADPS2...ADPS0, ADC prescaler
'                                                  =000: Div. factor 2
'                                                  =001: Div. factor 2
'                                                  =010: Div. factor 4
'                                                  =011: Div. factor 8
'                                                  =100: Div. factor 16
'                                                  =101: Div. factor 32
'                                                  =110: Div. factor 64
'                                                  =111: Div. factor 128
'SFIOR = &B00000000                                '&B00000000 is the default

On ADC ADC_isr                                    'ADC conversion complete interrupt
SREG.7 = 1                                        'Enable all interrupts

Cls
Cursor Off
Call ClearLCDLine(1 , bytChar)
LCD "SWR-Meter 2.01"
Call ClearLCDLine(2 , bytChar)
LCD "DL6GL 2014-03-06"
Wait 2
Cls

If Jmp_Coupler2 = 0 Then                          'Coupler2 selected
   LED_Coupler2 = 1                               'LED is lit
   bytUnit = 2
Else                                              'Coupler1 selected, default
   LED_Coupler2 = 0                               'LED is not lit
   bytUnit = 1
End If

'Initialize data --------------------------------------------------------------
$Include SWR_Bridge_Data_201.bas                  'Initialize Bridge data
Call CheckEEPROM                                  'Init/Read EEPROM data
sngADCCal = sngADCRef / 1024                      'ADC voltage calibration
sngADCFwd = 0                                     'ADC forward voltage
sngADCRev = 0                                     'ADC reverse voltage
bytADCRun = 0
bytADChannel = 0                                  'First ADC channel ADC0
bitADCdone = 0
sngSWRBarCal = bytBarLen * bytDots                'SWR bar dots (5 per character)
sngSWRBarCal = sngSWRBarCal / sngSWRBarMax        'SWR bar calibration
Waitms 500

'Part 1: Setup option ---------------------------------------------------
If Jmp_Setup = 0 Then                             'Setup jumper in
   Enable Timer0
   Timer0 = PresetTimer0                          'initialize Timer0
   bytMode = 1
   bitSetup = 0
   Call ClearLCDLine(1 , bytChar)
   LCD "Setup " ; bytMode ; "/7 Unit " ; bytUnit
   Cursor ON Blink

   Do

   'Check Mode Key ------------------------------------------------------------
   'Mode    switched consecutively by Mode key:
   '        (1) ADC reference voltage (units 1 and 2)
   '        The next items (2)...(7) may be set for 2 SWR couplers, e.g. QRP/QRO
   '        (2) Tandem match coupler toroid turns
   '        (3) Diode fit parameter A
   '        (4) Diode fit parameter B
   '        (5) Diode fit parameter C
   '        (6) Diode lower detection limit
   '        (7) Voltage calibration factor

      If bytKey_state.Key_mode = 1 Then           'Mode key pressed
         Incr bytMode                             'Next Mode
         bitSetup = 0
         bytPos = 1
         Call ClearLCDLine(1 , bytChar)
         LCD "Setup " ; bytMode ; "/7 Unit " ; bytUnit
      End If


      Select Case bytMode

         Case 1                                   'ADC reference voltage
            wrdAddr = 2
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngADCRef)
               strTmp10 = Fusing(sngADCRef , "#.###")
               Call ClearLCDLine(2 , bytChar)
               LCD "ADC-Vref: " ; strTmp10
               bytCursorFirst = 11
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 2                                   'Toroid turns
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 50
            Else                                  'Coupler1 selected, default
               wrdAddr = 6
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               ReadEeprom bytTurns , wrdAddr
               strTmp10 = Str(bytTurns)
               strTmp10 = Format(strTmp10 , "00")
               Call ClearLCDLine(2 , bytChar)
               LCD "Coupler Turns:" ; strTmp10
               bytCursorFirst = 15
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 1 , 1)           'Row 2, Byte, Multiplier 1
            End If

         Case 3                                   'Diode fit parameter A
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 51
            Else                                  'Coupler1 selected, default
               wrdAddr = 7
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngDiodeA)
               strTmp10 = Fusing(sngDiodeA , " #.#####")
               Call ClearLCDLine(2 , bytChar)
               LCD "DiodeA: " ; strTmp10
               bytCursorFirst = 9
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 4                                   'Diode fit parameter B
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 55
            Else                                  'Coupler1 selected, default
               wrdAddr = 11
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngDiodeB)
               strTmp10 = Fusing(sngDiodeB , " #.#####")
               Call ClearLCDLine(2 , bytChar)
               LCD "DiodeB: " ; strTmp10
               bytCursorFirst = 9
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 5                                   'Diode fit parameter C
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 59
            Else                                  'Coupler1 selected, default
               wrdAddr = 15
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngDiodeC)
               strTmp10 = Fusing(sngDiodeC , " #.#####")
               Call ClearLCDLine(2 , bytChar)
               LCD "DiodeC: " ; strTmp10
               bytCursorFirst = 9
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 6                                   'Diode lower detection limit
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 63
            Else                                  'Coupler1 selected, default
               wrdAddr = 19
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngDiodeLow)
               strTmp10 = Fusing(sngDiodeLow , "#.###")
               Call ClearLCDLine(2 , bytChar)
               LCD "Diode-Low: " ; strTmp10
               bytCursorFirst = 12
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 7                                   'Voltage calibration factor
            If Jmp_Coupler2 = 0 Then              'Coupler2 selected
               wrdAddr = 67
            Else                                  'Coupler1 selected, default
               wrdAddr = 23
            End If
            If bitSetup = 0 Then                  'Prepare data
               bitSetup = 1
               Call EEPROMReadSng(sngCalUnit)
               strTmp10 = Fusing(sngCalUnit , "#.###")
               Call ClearLCDLine(2 , bytChar)
               LCD "CalFactor: " ; strTmp10
               bytCursorFirst = 12
               Locate 2 , bytCursorFirst
            Else                                  'Customize data
               Call SetValue(2 , 3 , 1)           'Row 2, Single, Multiplier 1
            End If

         Case 8                                   'From Mode 7 back to 1
            Call ClearLCDLine(2 , bytChar)
            LCD "Setup finished  "
            bytMode = 0
            bitSetup = 0
            bytPos = 1
      End Select

      Waitms 200                                  'wait for key released
      bytKey_press = 0                            'reset Key_press

   Loop

   'Part 2: SWR/power display -------------------------------------------------
Else                                              'Customizing jumper out

   ADCSRA.ADSC = 1                                'Start ADC
   Do

      If Jmp_Power = 0 Then                       'Jmp_Power in: eff. power measured
         LED_Power = 1                            'Turn LED_Power on, eff. power
      Else
         LED_Power = 0                            'Turn LED_Power off, peak power
      End If

      Call GetADCData                             'Accumulate ADC values
      If bitADCPWR = 1 Then                       'ADC data accumulated
         bytADCRun = 0                            'Initialize next ADC run
         Call ADCalibrate                         'Calibrate ADC values
         Select Case bytPWRerr
            Case 0                                'OK, calculate SWR
               Call ShowPower                     'Show power
               Call ShowSWR                       'Show SWR in row bytRowSWR
            Case 1
               Locate bytRowSWR , 1
               LCD "Err: Signal low "
            Case 2
               Locate bytRowSWR , 1
               LCD "Err: FWD < REV  "
            Case 3
               Locate bytRowPwr , 1
               LCD "No RF signal    "
         End Select
         sngADCFwd = 0                            'Initialize for next ADC run
         sngADCRev = 0
      End If

   Loop

End If

End

'==============================================================================
Timer0_isr:
'Interrupt Service Routine for Timer0 (Keys)
'got from Thomas Baer in microcontroller.net
'Original (ASM) by Peter Dannegger
'http://www.mikrocontroller.net/topic/tasten-entprellen-bulletproof
'Output: bytKey_press bit pattern, key which changed from open to pressed
'        bytKey_state bit pattern, Key state, 1:pressed or 0:not pressed
'                     bits 0...8: keys 0...8 @ Key_port
'        Set bytKey_press = 0 in the main program after processing key data

   Timer0 = Presettimer0
   'Keys ----------------------------------------------------------------------
   'got from Thomas Baer in Microcontroller.net
   bytKey_save = bytKey_state                     'save old status before keys were pressed
   bytIwr0 = Key_port                             'get Bits of the KeyPorts
   bytIwr0 = Not bytIwr0
   bytIwr0 = bytIwr0 Xor bytKey_state
   bytKey_ct0 = bytKey_ct0 And bytIwr0
   bytKey_ct1 = bytKey_ct1 And bytIwr0
   bytKey_ct0 = Not bytKey_ct0
   bytKey_ct1 = bytKey_ct1 Xor bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct1
   bytKey_state = bytKey_state Xor bytIwr0
   bytIwr0 = bytIwr0 And bytKey_state
   bytKey_press = bytKey_press Or bytIwr0

Return

'==============================================================================
Function KeyPressLong(byVal KeyNum As Byte) As Byte
'Check if Key Key_Port.KeyNum is pressed short or long
'KeyPressLong = 1: pressed long (~ 1 sec), = 0: short

   bytTimer0_runs = 0
   Do                                             'Wait for key to be released
      Incr bytTimer0_runs
      Waitms 10
   Loop Until bytKey_state.KeyNum = 0 Or bytTimer0_runs >= KeyCounter

   If bytTimer0_runs >= KeyCounter Then
      KeyPressLong = 1
      bytTimer0_runs = 0                          'key was pressed long
   Else
      KeyPressLong = 0                            'key was pressed short
   End If
End Function

'==============================================================================
ADC_isr:
   'Interrupt Service Routine for ADC, conversion complete
   'Read 10bit ADC Data Registers ADCL & ADCH
   'bytADCLo and bytADCHi are overlayed with wrdADC
   'Set next MUX channel and start ADC again

   bytADCLo = ADCL                                'Read ADCL first...
   bytADCHi = ADCH                                '... then ADCH
   bytADCValIdx = bytADChannel + 1                'Next value array position
   wrdADCVal(bytADCValIdx) = wrdADC               '10bit ADC value
   Incr bytADChannel                              'Next MUX channel
   If bytADChannel > bytADCUsed Then
      bytADChannel = 0
      bitADCdone = 1                              'Accumulation complete
   End If
   Admux = bytADMUX Or bytADChannel               'Set next MUX channel
   ADCSRA.ADSC = 1                                'Start ADC again
Return

'==============================================================================
Sub GetADCData
'Get accumulated ADC Data
'Input:  wrdADCVal(1)   ADC forward voltage
'        wrdADCVal(2)   ADC reverse voltage
'        bytADCRep      Number of repetitive ADC accumulations
'Output: sngADCFwd      Mean ADC forward voltage
'        sngADCRev      Mean ADC reverse voltage
'        bytADCRun      Current ADC accumulation number

   If bitADCdone = 1 Then                         'ADC accumulation complete
      If bytADCRun < bytADCRep Then               'Accumulate bytADCRep times
         sngADCFwd = sngADCFwd + wrdADCVal(1)     'ADC RF peak power value
         sngADCRev = sngADCRev + wrdADCVal(2)     'ADC Supply voltage
         Incr bytADCRun
         bitADCPWR = 0                            'ADC accumulating
      Else                                        'Mean values complete
         sngFwd = sngADCFwd / bytADCRep           'Mean values
         sngRev = sngADCRev / bytADCRep
         bitADCPWR = 1                            'ADC accumulation done
      End If
      bitADCdone = 0
   End If

End Sub

'==============================================================================
Sub ADCalibrate
'Calibrate ADC for power/SWR measurement
'Input:  sngFWD      Forward ADC count
'        sngREV      Reverse ADC count
'        sngDiodeLow Diode threshold value
'Output  sngFWD      Forward calibrated voltage
'        sngREV      Reverse calibrated voltage
'        bitTXOn     = ON: TX is on, = OFF: TX is off
'        bytPWRerr   Error flag
'                    = 0: no error
'                    = 1: signal low
'                    = 2: FWD < REV
'                    = 3: no signal
'Function DiodeResponse for correction of diode response

   If sngFWD > 5 Then                             'TX is on (Fwd >50 mV)
      If sngFWD > sngREV Then                     'FWD must be > REV
         sngFWD = sngFWD * sngADCCal              'Forward peak voltage
         sngREV = sngREV * sngADCCal              'Reverse peak voltage
         If sngFWD > sngDiodeLow Then             'Signal level OK
            sngFWD = DiodeResponse(sngFWD)        'Diode correction forward
            sngREV = DiodeResponse(sngREV)        'Diode correction reverse
            bytPWRerr = 0                         'Calibration done
         Else
            bytPWRerr = 1                         'Signal low
         End If
      Else
         bytPWRerr = 2                            'FWD < REV
      End If
   Else
      bytPWRerr = 3                               'Fwd < 25 mV, TX must be off
   End If

End Sub

'==============================================================================
Function DiodeResponse(byVal sngDC As Single) As Single
'Calculate "real" HF voltage from measured DC voltage sngDC
'Formula valid for Schotty BAT43 diode peak detector
'Correction factor:  CF = a * DC^b + c
'"Real" HF:          HF = CF * DC voltage
'Input:  sngDC, Diode DC voltage (Forward or Reverse)
'        sngDiodeA, ...B, ...C, best fit parameters for correction function
'Output: HF voltage from best fit correction function

   sngTmp0 = sngDiodeB                            'Const is not accepted by Power
   sngTmp1 = Power(sngDC , sngTmp0)
   sngTmp1 = sngTmp1 * sngDiodeA
   sngTmp1 = sngTmp1 + sngDiodeC                  'Correction factor
   DiodeResponse = sngDC * sngTmp1                'Multiply with DC voltage

End Function

'==============================================================================
Sub ShowPower
'Calculate an show forward & reverse RF Power
'Power   = (voltage)^2 / impedance
'eff voltage  = peak voltage / SQR(2)
'Input:  sngFWD      forward peak voltage
'        sngREV      reverse peak voltage
'        sngImp      circuit impedance
'        sngTurns    Voltage calibration factor (toroid turns ratio)
'        sngCalUnit  Voltage calibration factor (Unit1: 1, unit2: ?)

   sngTmp0 = sngFWD * sngCalUnit                  'Peak forward voltage
   sngTmp1 = sngREV * sngCalUnit                  'Peak reverse voltage
   If Jmp_Power = 0 Then                          'Jumper in: effective power
      sngTmp0 = 0.7071 * sngTmp0                  'eff. forward voltage
      sngTmp1 = 0.7071 * sngTmp1                  'eff. reverse voltage
   End If
   sngTmp0 = sngTmp0 * bytTurns                   'Voltage calibration
   sngTmp1 = sngTmp1 * bytTurns
   sngTmp0 = sngTmp0 * sngTmp0                    'forward voltage squared
   sngTmp1 = sngTmp1 * sngTmp1                    'reverse voltage squared
   sngTmp0 = sngTmp0 / sngImp                     'forward power
   sngTmp1 = sngTmp1 / sngImp                     'reverse power

   wrdTmp0 = sngTmp0 * 10                         'forward power, show 1 decimal
   wrdTmp1 = sngTmp1 * 10                         'reverse power, show 1 decimal

   strTmp6 = Str(wrdTmp0)
   strTmp6 = Format(strTmp6 , "  0.0")
   strDis = "F " + strTmp6                        'forward power ...
   strTmp6 = Str(wrdTmp1)
   strTmp6 = Format(strTmp6 , "  0.0")
   strDis = strDis + " R " + strTmp6 + "W"        '... and reverse power
   Locate bytRowPwr , 1
   LCD strDis                                     'Show me

 End Sub

'==============================================================================
Sub ShowSWR
'Paint value sngVal on LCD, Row bytRow, as bar graph
'Input:  sngFWD         Foreward voltage
'        sngREV         Reverse  voltage
'        sngSWRMax      max. SWR to be displayed (end of bar)
'        sngSWRBarCal   calibration factor (dots / max value)

   sngTmp0 = sngFWD + sngREV
   sngTmp1 = sngFWD - sngREV
   sngSWR = sngTmp0 / sngTmp1                     'SWR
   If sngSWR > 9.99 Then
      sngSWR = 9.99                               'SWR display limit
   End If
   If sngSWR > sngSWRMax Then                     'Max SWR exeeded?
      LED_SWR = 1                                 'Yes, LED_SWR on
   Else
      LED_SWR = 0                                 'No, LED_SWR off
   End If

   sngTmp0 = sngSWR
   If sngTmp0 > sngSWRBarMax Then
      sngTmp0 = sngSWRBarMax                      'Display limit
   End If
   sngTmp0 = sngSWRBarCal * sngTmp0               'Bar calibration
   wrdTmp2 = Int(sngTmp0)
   sngTmp0 = sngTmp0 / bytDots
   wrdTmp0 = Int(sngTmp0)                         'Number of 5 dot full blocks
   wrdTmp1 = wrdTmp0 * bytDots
   wrdTmp1 = wrdTmp2 - wrdTmp1                    'Remaining partial bars

   strDis = ""
   If wrdTmp0 > 0 Then                            'Handle full 5 dot blocks
      strDis = String(wrdTmp0 , 4)                'Fill with Chr(4)
   End If

   Select Case wrdTmp1                            'Add last bar(s) < 5 dots
      Case 1
           strDis = strDis + Chr(0)               '1 bar
      Case 2
           strDis = strDis + Chr(1)               '2 bars
      Case 3
           strDis = strDis + Chr(2)               '3 bars
      Case 4
           strDis = strDis + Chr(3)               '4 bars
   End Select

   wrdTmp1 = bytBarLen - 4                        'Space for Tag 'SWR'?
   If wrdTmp0 < wrdTmp1 Then
      bytTmp0 = 1                                 'yes
   Else
      bytTmp0 = 0                                 'no
   End If

   wrdTmp0 = Len(strDis)                          'string length up to now
   wrdTmp1 = bytChar - wrdTmp0                    'remaining characters
   wrdTmp1 = wrdTmp1 - 4                          '4 char's for value
   If wrdTmp1 > 0 Then
      strDis = strDis + Space(wrdTmp1)            'Fill with spaces
   End If
   If bytTmp0 = 1 Then                            'Add tag 'SWR'
      wrdTmp0 = Len(strDis)
      wrdTmp1 = wrdTmp0 - 4
      strDis = Left(strDis , wrdTmp1)
      strDis = strDis + "SWR "
   End If

   wrdTmp0 = sngSWR * 100                         'show 2 decimals
   strTmp6 = str(wrdTmp0)
   strTmp6 = Format(strTmp6 , "0.00")
   strDis = strDis + strTmp6                      'Add sngSWR value at the end
   Locate bytRowSWR , 1
   LCD strDis                                     'Show me

End Sub

'==============================================================================
Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
'Clear LCD bytChars characters in line bytRow
   Locate bytRow , 1
   LCD SPC(bytChars)
   Locate bytRow , 1
End Sub

'==============================================================================
Sub SetValue(byVal bytRow As Byte , byVal bytType As Byte , byVal intMult As Integer)
'Set EEPROM value via keys and restore it to EEPROM
'Digits are selected via right/left keys
'Numbers are changed via up-key (0...9)
'Input:  bytRow                  LCD row to display value
'        bytType                 Type of value
'                                1 = Byte
'                                2 = Word
'                                3 = Single
'        intMult                 Multiplicator/divisor for value
'                                +10/+100/+1000: multiply by 10/100/1000
'                                -10/-100/-1000: divide   by 10/100/1000
'        strTmp10                formatted value to be set (String * 10)
'                                overlayed with bytTmp10(1...10)
'        bytPos                  First position in strTmp10
'        wrdAddr                 EEPROM address to store value
'        bytCursorPos            Cursor start position (digit)
'                                must be set to first LCD digit

   bytNumChar = Len(strTmp10)                     'characters including sign
   bytPointPos = INSTR(1 , strTmp10 , ".")        'Decimal point position
   bytSignPos = INSTR(1 , strTmp10 , "-")         'Minus sign position
   If bytPos = bytSignPos Then bytPos = bytPos + 1       'Skip sign

   'Select digit to the right (Key_Right)
   'Key_Right pressed short: select digit
   'Key_Right pressed long : store value to EEPROM
   If bytKey_state.Key_Right = 1 Then             'Key_Right pressed = up
      bytLong = KeyPressLong(Key_Right)
      If bytLong = 0 Then                         'Pressed short
         If bytPos < bytNumChar Then
            Incr bytPos
            If bytPos = bytPointPos Then          'Is decimal point
               Incr bytPos                        'skip decimal point
            End If
         End If
      Else                                        'Pressed long: store
         Select Case bytType
            Case 1                                'Byte value
               sngTmp0 = Val(strTmp10)
               If intMult > 0 Then
                  sngTmp0 = sngTmp0 * intMult
               ElseIf intMult < 0 Then
                  sngTmp0 = sngTmp0 / intMult
               End If
               bytTmp0 = sngTmp0 + 0.5
               WriteEEprom bytTmp0 , wrdAddr
            Case 2                                'Word value
               sngTmp0 = Val(strTmp10)
               If intMult > 0 Then
                  sngTmp0 = sngTmp0 * intMult
               ElseIf intMult < 0 Then
                  sngTmp0 = sngTmp0 / intMult
               End If
               wrdTmp0 = sngTmp0 + 0.5
               Call EEPROMWriteWrd(wrdTmp0)
            Case 3                                'Single value
               sngTmp0 = Val(strTmp10)
               Call EEPROMWriteSng(sngTmp0)
         End Select

         Call ClearLCDLine(1 , 16)
         LCD strTmp10 ; " stored"
      End If
   End If

   'Select digit downwards (Key_Left)
   If bytKey_state.Key_Left = 1 Then              'Key_Left pressed = down
      If bytPos > 1 Then
         Decr bytPos
         If bytPos = bytPointPos Then             'Is decimal point
            Decr bytPos                           'skip decimal point
         End If
         If bytPos = bytSignPos Then              'Is minus sign
            Incr bytPos                           'Skip minus sign
         End If
      End If
   End If
   bytCursorPos = bytCursorFirst + bytPos         'Cursor position
   bytCursorPos = bytCursorPos - 1
   Locate bytRow , bytCursorPos

   'Change selected digit via Key_Up
   If bytKey_state.Key_Up = 1 Then                'Key_up is pressed
      bytSave = bytTmp10(bytPos)                  'Char (ASCII) to be changed
      If bytSave < 48 Or bytSave >= 57 Then       'ASCII 0 or 9
         bytSave = 48                             'ASCII 0
      Else
         Incr bytSave
      End If
      strTmp1 = Chr(bytSave)                      'Changed value (ASCII bytSave)
      bytTmp10(bytPos) = strTmp1                  '... into string strTmp10
      LCD strTmp1
   End If

End Sub

'==============================================================================
Sub CheckEEPROM
'Check if SWR bridge data are stored in EEPROM
'if not, write bridge data
'Read EEPROM data
'In a new controller EEPROM data = 255
'SWR bridge data:
'Byte Variable    Type     Description
' 0   --          Byte     Not used
' 1   --          Byte     = 1: data are stored
' 2   bytTurns    Byte     Number of tandem match toroid turns
' 3   sngADCRef  Single   ADC reference voltage (V)
' 7   sngDiodeA   Single   Diode fit parameter A for y = A * x^B + C
'11   sngDiodeB   Single   Diode fit parameter B for y = A * x^B + C
'15   sngDiodeC   Single   Diode fit parameter C for y = A * x^B + C
'19   sngDiodeLow Single   Diode lower detection limit (V)

   ReadEeprom bytTmp1 , 1                         'Check Address #1

   If bytTmp1 = 1 Then                            'Read available EEPROM data
      Call ClearLCDLine(1 , bytChar)
      LCD "Read EEPROM..."
      wrdAddr = 2
      Call EEPROMReadSng(sngADCRef)
      If Jmp_Coupler2 = 1 Then                    'Coupler2 not selected
         wrdAddr = 6                              'Start address coupler1 data
      Else
         wrdAddr = 50                             'Start address coupler2 data
      End If
      ReadEEPROM bytTurns , wrdAddr
      Incr wrdAddr
      Call EEPROMReadSng(sngDiodeA)
      Call EEPROMReadSng(sngDiodeB)
      Call EEPROMReadSng(sngDiodeC)
      Call EEPROMReadSng(sngDiodeLow)
      Call EEPROMReadSng(sngCalUnit)
      Call ClearLCDLine(2 , bytChar)
      Lcd "... Read done"
      Wait 1
   Else                                           'Is a brandnew controller
      Call ClearLCDLine(1 , bytChar)              'Initialize it with data...
      LCD "Init EEPROM..."                        '...form SWR_Bridge_data.bas
      wrdAddr = 2
      Call EEPROMWriteSng(sngADCRef)
      For wrdTmp0 = 1 To 2                        'Write coupler 1 & 2 data
         If wrdTmp0 = 1 Then
            wrdAddr = 6                           'Start address coupler1 data
         Else
            wrdAddr = 50                          'Start address coupler2 data
         End If
         WriteEEPROM bytTurns , wrdAddr
         Incr wrdAddr
         Call EEPROMWriteSng(sngDiodeA)
         Call EEPROMWriteSng(sngDiodeB)
         Call EEPROMWriteSng(sngDiodeC)
         Call EEPROMWriteSng(sngDiodeLow)
         Call EEPROMWriteSng(sngCalUnit)
      Next wrdTmp0
      bytTmp0 = 1
      wrdAddr = 1
      WriteEEPROM bytTmp0 , wrdAddr
      Call ClearLCDLine(2 , bytChar)
      Lcd "... Write done"
      Wait 1
   End If

End Sub

'==============================================================================
Sub EEPROMWriteWrd(byRef wrdWord As Word)
   'Decode wrdWord to 2 bytes of bytWrdEEP for writing to EEPROM
   'Input:  wrdWord to be decoded & written to EEPROM
   '        wrdAddr first EEPROM address

   wrdEEPROM = wrdWord                            'wrdEEPROM overlayed @ bytWrdEEP
   For bytTmp0 = 1 To 2
      WriteEEPROM bytWrdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0
End Sub

'==============================================================================
Sub EEPROMReadWrd(byRef wrdWord As Word)
   'Encode 2 bytes read from EEPROM to wrdWord
   'Input:  wrdAddr first EEPROM address
   'Output: wrdWord encoded unsigned integer (word) variable

   For bytTmp0 = 1 To 2
      ReadEEPROM bytWrdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0
   wrdWord = wrdEEPROM                            'wrdEEPROM overlayed @ bytWrdEEP
End Sub

'==============================================================================
Sub EEPROMWriteSng(byRef sngSingle As Single)
   'Decode sngSingle to 4 bytes of bytSngEEP for writing to EEPROM
   'Input sngSingle is set to sngEEPROM, overlayed with bytSngEEP(i), i=1...4
   'The 4 bytes bytSngEEP(i)are written to EEPROM
   'Input:  sngSingle to be decoded & written to EEPROM
   '        wrdAddr first EEPROM address


   sngEEPROM = sngSingle                          'Set input
   For bytTmp0 = 1 To 4
      WriteEEPROM bytSngEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0
End Sub

'==============================================================================
Sub EEPROMReadSng(byRef sngSingle As Single)
   'Encode 4 bytes read from EEPROM to sngSingle
   'The 4 bytes are read to bytSngEEP(i), i=1...4
   'bytSngEEP is overlayed @ sngEEPROM
   'sngEEPROM is set to output sngSingle
   'Input:  wrdAddr first EEPROM address
   'Output: sngSingle encoded to single variable read from EEPROM


   For bytTmp0 = 1 To 4
      ReadEEPROM bytSngEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0
   sngSingle = sngEEPROM                          'Set output
End Sub